MOBILE_USERAGENT = 'Reddit/Version 2020.47.0/Build 302202/Android 11'
WEB_USERAGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0'
OAUTH_CLIENT_ID_B64 = "b2hYcG9xclpZdWIxa2c6"
USER_AGENT = 'Jand/3.0.82'
SB_User_Agent = 'Android%2Fc3.0.144F'
SB_ai = '2515BDA8-9D3A-47CF-9325-330BC37ADA13'
SB_PROXY_CHATMEDIA = 'https://sendbirdproxyk8s.chat.redditmedia.com'
S_REDDIT = 'https://s.reddit.com'
OAUTH_REDDIT = 'https://oauth.reddit.com'
WWW_REDDIT = 'https://old.reddit.com'
ACCOUNTS_REDDIT = 'https://accounts.reddit.com'

# ws payloads
MESG_regular = 'MESG{{"channel_url":"{channel_url}","message":"{text}","data":"{{\\"v1\\":{{\\"preview_collapsed\\":false,\\"embed_data\\":{{}},\\"hidden\\":false,\\"highlights\\":[],\\"message_body\\":\\"{text}\\"}}}}","mention_type":"users","req_id":"{req_id}"}}\n'
MESG_snoo = 'MESG{{"channel_url":"{channel_url}","message":"","data":"{{\\"v1\\":{{\\"preview_collapsed\\":false,\\"embed_data\\":{{\\"site_name\\":\\"Reddit\\"}},\\"hidden\\":false,\\"snoomoji\\":\\"{snoomoji}\\"}}}}","mention_type":"users","req_id":"{req_id}"}}\n'
TPST = 'TPST{{"channel_url":"{channel_url}","time":{time},"req_id":""}}\n'
TPEN = 'TPEN{{"channel_url":"{channel_url}","time":{time},"req_id":""}}\n'



import requests
from .._utils.consts import SB_PROXY_CHATMEDIA, S_REDDIT, USER_AGENT, SB_User_Agent, SB_ai, WEB_USERAGENT, WWW_REDDIT
import json
from .models import Channel, Message


def _get_user_id(username):
    response = requests.get(f"{WWW_REDDIT}/user/{username}/about.json",
                            headers={'user-agent': WEB_USERAGENT}).json()
    u_id = response.get('data', {}).get('id')
    if u_id is None:
        return None
    else:
        return f't2_{u_id}'


class Tools:
    def __init__(self, reddit_auth):
        self._req_sesh = requests.Session()
        self._req_sesh.headers = {
            'User-Agent': USER_AGENT,
            'SendBird': f'Android,29,3.0.82,{SB_ai}',
            'SB-User-Agent': SB_User_Agent
        }
        self._reddit_auth = reddit_auth
        try:
            _ = self._reddit_auth.reddit_username
            self._is_reauthable = True
        except AttributeError:
            self._is_reauthable = False

    def _handled_req(self, method, uri, **kwargs):
        while True:
            response = self._req_sesh.request(method, uri, **kwargs)
            if response.status_code == 401 and self._is_reauthable:
                new_access_token = self._reddit_auth.refresh_api_token()
                new_headers = kwargs.get('headers', {})
                new_headers.update({'Authorization': f'Bearer {new_access_token}'})
                kwargs.update({'headers': new_headers})
                continue
            else:
                return response

    def delete_message(self, channel_url, msg_id, session_key):
        uri = f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/messages/{msg_id}"
        self._handled_req(method='DELETE', uri=uri, headers={'Session-Key': session_key})

    def kick_user(self, channel_url, user_id, duration):
        data = json.dumps({
            'channel_url': channel_url,
            'user_id': user_id,
            'duration': duration
        })
        uri = f'{S_REDDIT}/api/v1/channel/kick/user'
        self._handled_req(method='POST', uri=uri, headers={'Authorization': f'Bearer {self._reddit_auth.api_token}'},
                          data=data)

    def invite_user(self, channel_url, nicknames):
        assert isinstance(nicknames, list)
        users = []
        for nickname in nicknames:
            users.append({'user_id': _get_user_id(nickname), 'nickname': nickname})
        data = json.dumps({
            'users': users
        })
        url = f'{S_REDDIT}/api/v1/sendbird/group_channels/{channel_url}/invite'
        self._handled_req(method='POST', uri=url, headers={'Authorization': f'Bearer {self._reddit_auth.api_token}'},
                          data=data)

    def accept_chat_invite(self, channel_url, session_key):
        data = json.dumps({
            'user_id': self._reddit_auth.user_id
        })
        url = f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/accept'
        self._handled_req(method='PUT', uri=url, headers={'Session-Key': session_key}, data=data)

    def get_channels(self, limit, order, show_member, show_read_receipt, show_empty, member_state_filter, super_mode,
                     public_mode, unread_filter, hidden_mode, show_frozen, session_key):
        params = {
            'limit': limit,
            'order': order,
            'show_member': show_member,
            'show_read_receipt': show_read_receipt,
            'show_delivery_receipt': 'true',
            'show_empty': show_empty,
            'member_state_filter': member_state_filter,
            'super_mode': super_mode,
            'public_mode': public_mode,
            'unread_filter': unread_filter,
            'hidden_mode': hidden_mode,
            'show_frozen': show_frozen,
        }
        url = f'{SB_PROXY_CHATMEDIA}/v3/users/{self._reddit_auth.user_id}/my_group_channels'
        response = self._handled_req(method='GET', uri=url, headers={'Session-Key': session_key}, params=params)
        return [Channel.from_dict(channel, infer_missing=True) for channel in response.json()['channels']]

    def leave_chat(self, channel_url, session_key):
        data = json.dumps({
            'user_id': self._reddit_auth.user_id
        })
        url = f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/leave'
        self._handled_req(method='PUT', uri=url, headers={'Session-Key': session_key}, data=data)

    def create_channel(self, nicknames, group_name, own_name):
        users = [{"user_id": self._reddit_auth.user_id, "nickname": own_name}]
        for nickname in nicknames:
            users.append({'user_id': _get_user_id(nickname), 'nickname': nickname})
        data = json.dumps({
            'users': users,
            'name': group_name
        })
        url = f'{S_REDDIT}/api/v1/sendbird/group_channels'
        response = self._handled_req(method='POST', uri=url,
                                     headers={'Authorization': f'Bearer {self._reddit_auth.api_token}'},
                                     data=data)
        return Channel.from_json(response.text, infer_missing=True)

    def hide_chat(self, user_id, channel_url, hide_previous_messages, allow_auto_unhide, session_key):
        data = json.dumps({
            'user_id': user_id,
            'hide_previous_messages': hide_previous_messages,
            'allow_auto_unhide': allow_auto_unhide
        })
        self._handled_req(method='PUT', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/hide',
                          headers={'Session-Key': session_key}, data=data)

    def get_older_messages(self, channel_url, prev_limit, next_limit, reverse, session_key):
        params = {
            'is_sdk': 'true',
            'prev_limit': prev_limit,
            'next_limit': next_limit,
            'include': 'false',
            'reverse': reverse,
            'message_ts': 9007199254740991,
            'with_sorted_meta_array': 'false',
            'include_reactions': 'false',
            'include_thread_info': 'false',
            'include_replies': 'false',
            'include_parent_message_text': 'false'
        }

        response = self._handled_req(method='GET', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/messages',
                                     headers={'Session-Key': session_key}, params=params)
        return [Message.from_dict(msg, infer_missing=True) for msg in response.json()['messages']]



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Dict, Optional, List
from enum import Enum


@dataclass_json
@dataclass(frozen=True)
class _Channel:
    name: str
    member_count: int
    custom_type: str
    channel_url: str
    created_at: int
    # cover_url: str
    max_length_message: int
    data: str


@dataclass_json
@dataclass(frozen=True)
class _MentionedUser:
    nickname: str
    # metadata: dict
    # require_auth_for_profile_image: bool
    # profile_url: str
    user_id: str


@dataclass_json
@dataclass(frozen=True)
class _User:
    require_auth_for_profile_image: bool
    is_active: Optional[bool]
    is_blocked_by_me: bool
    role: str
    user_id: str
    nickname: str
    # profile_url: str
    # metadata: dict


@dataclass_json
@dataclass(frozen=True)
class _CreatedBy:
    # require_auth_for_profile_image: bool
    nickname: str
    user_id: str
    # profile_url: str


@dataclass_json
@dataclass(frozen=True)
class Message:
    # message_survival_seconds: int
    # custom_type: str
    mentioned_users: List[_MentionedUser]
    # translations: dict
    updated_at: int
    is_op_msg: bool
    is_removed: bool
    user: _User
    # file: dict
    message: str
    data: str
    # message_retention_hour: int
    # silent: bool
    type: str
    created_at: int
    req_id: str
    mention_type: str
    channel_url: str
    message_id: int


@dataclass_json
@dataclass(frozen=True)
class _Member:
    is_blocking_me: Optional[bool]
    user_id: str
    is_muted: bool
    friend_name: Optional[str]
    is_active: Optional[bool]
    is_blocked_by_me: bool
    state: str
    # friend_discovery_key: null
    role: str
    is_online: bool
    require_auth_for_profile_image: bool
    # last_seen_at: int
    nickname: str
    # profile_url:str
    # metadata: dict


class CustomType(Enum):
    group = 'group'
    direct = 'direct'


class MemberState(Enum):
    joined = 'joined'
    invited = 'invited'


@dataclass_json
@dataclass(frozen=True)
class Channel:
    invited_at: int
    custom_type: CustomType
    # is_ephemeral: bool
    read_receipt: Dict[str, int]
    member_state: MemberState
    freeze: bool
    created_by: Optional[_CreatedBy]
    is_hidden: bool
    # disappearing_message: dict
    is_push_enabled: bool
    joined_ts: Optional[int]
    is_created: Optional[bool]
    member_count: int
    # my_role: str
    # is_broadcast: bool
    last_message: Optional[Message]
    user_last_read: int
    unread_mention_count: int
    # sms_fallback: dict
    # is_discoverable: bool
    # ignore_profanity_filter: bool
    channel_url: str
    # operators: list
    channel: _Channel
    # message_survival_seconds: int
    unread_message_count: int
    # is_distinct: bool
    # is_muted: bool
    # hidden_state: str
    cover_url: str
    members: List[_Member]
    is_public: bool
    # data: str
    # ts_message_offset: int
    joined_member_count: int
    is_super: bool
    name: str
    created_at: int
    # is_access_code_required: bool
    # push_trigger_option: str
    max_length_message: int
    inviter: _CreatedBy
    count_preference: str



import json
from types import SimpleNamespace
from enum import Enum


class FrameType(Enum):
    MESG = 'MESG'
    SYEV = 'SYEV'
    DELM = 'DELM'
    TPEN = 'TPEN'
    TPST = 'TPST'
    READ = 'READ'
    USEV = 'USEV'
    MACK = 'MACK'
    BRDM = 'BRDM'
    LOGI = 'LOGI'


class FrameModel(SimpleNamespace):
    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "({})".format(", ".join(items))


def convert_to_framemodel(d):
    return json.loads(d, object_hook=lambda d: FrameModel(**d))


def get_frame_data(data):
    data_r = data[4:]
    data_j = convert_to_framemodel(data_r)
    type_f = FrameType[data[:4]]

    # some presets
    if type_f == FrameType.MESG and data_j.message == "":
        wfdata = convert_to_framemodel(data_j.data)
        data_j.message = wfdata.v1.snoomoji
    data_j.type_f = type_f

    return data_j



import websocket
from ._utils.rate_limiter import RateLimiter
import time
from ._utils.frame_model import get_frame_data, FrameType
import logging
from threading import Thread
from ._utils import ws_utils
from ._utils.consts import MESG_regular, MESG_snoo, TPST, TPEN, MOBILE_USERAGENT


logging.basicConfig(level=logging.INFO, datefmt='%H:%M', format='%(asctime)s, %(levelname)s: %(message)s')


def _print_chat_(resp, channelid_sub_pairs):
    if resp.type_f == FrameType.MESG:
        print(f"{resp.user.name}@{channelid_sub_pairs.get(resp.channel_url)}: {resp.message}")


class WebSockClient:
    def __init__(self, access_token, user_id, enable_trace=False, print_chat=True, log_websocket_frames=False,
                 other_logging=True, global_blacklist_users=None, global_blacklist_words=None):
        self._user_id = user_id
        if global_blacklist_words is None:
            global_blacklist_words = set()
        if global_blacklist_users is None:
            global_blacklist_users = set()
        assert isinstance(global_blacklist_words, set), "blacklists must be set()s"
        assert isinstance(global_blacklist_users, set), "blacklists must be set()s"
        self.global_blacklist_words = global_blacklist_words
        self.global_blacklist_users = global_blacklist_users

        self.channelid_sub_pairs = {}
        self.RateLimiter = RateLimiter

        self.logger = logging.getLogger(__name__)
        self.logger.disabled = not other_logging
        websocket.enableTrace(enable_trace)

        ws_url = ws_utils.get_ws_url(self._user_id, access_token)
        self.ws = self._get_ws_app(ws_url)

        self.req_id = int(time.time() * 1000)
        self.own_name = None
        self.print_chat = print_chat
        self.log_websocket_frames = log_websocket_frames
        self.last_err = None
        self.is_logi_err = False
        self.session_key = None

        self._get_current_channels = None

        self.after_message_hooks = []

    def _get_ws_app(self, ws_url):
        ws = websocket.WebSocketApp(ws_url,
                                    on_message=lambda ws, msg: self.on_message(ws, msg),
                                    on_error=lambda ws, msg: self.on_error(ws, msg),
                                    on_close=lambda ws: self.on_close(ws),
                                    on_open=lambda ws: self.on_open(ws),
                                    header={'User-Agent': MOBILE_USERAGENT}
                                    )
        return ws

    def ws_run_forever(self, skip_utf8_validation):
        self.ws.run_forever(ping_interval=15, ping_timeout=5, skip_utf8_validation=skip_utf8_validation)

    def update_ws_app_urls_access_token(self, access_token):
        self.ws.url = ws_utils.get_ws_url(self._user_id, access_token)

    def on_open(self, ws):
        self.logger.info("### successfully connected to the websocket ###")

    def on_message(self, ws, message):
        resp = get_frame_data(message)
        if self.print_chat:
            _print_chat_(resp, self.channelid_sub_pairs)
        if self.log_websocket_frames:
            self.logger.info(message)
        if resp.type_f == FrameType.LOGI:
            self.logger.info(message)
            self._logi(resp)

        if resp.type_f == FrameType.MESG and resp.user.name in self.global_blacklist_users:
            return

        Thread(target=self._response_loop, args=(resp,), daemon=True).start()

    def _logi(self, resp):
        try:
            logi_err = resp.error
        except AttributeError:
            logi_err = None
        if logi_err is None:
            self.session_key = resp.key
            self.current_channels = self._get_current_channels(limit=100, order='latest_last_message', show_member=True,
                                                               show_read_receipt=True, show_empty=True,
                                                               member_state_filter='joined_only', super_mode='all',
                                                               public_mode='all', unread_filter='all',
                                                               hidden_mode='unhidden_only', show_frozen=True,
                                                               session_key=self.session_key)
            self.update_channelid_sub_pair()
            self.own_name = resp.nickname
        else:
            self.logger.error(f"err: {resp.message}")
            self.is_logi_err = True

    def update_channelid_sub_pair(self):
        self.channelid_sub_pairs = ws_utils.pair_channel_and_names(channels=self.current_channels,
                                                                   user_id=self._user_id)

    def _response_loop(self, resp):
        for func in self.after_message_hooks:
            if func(resp):
                break

    def ws_send_message(self, text, channel_url):
        if self.RateLimiter.is_enabled and self.RateLimiter.check():
            return
        if any(blacklist_word in text.lower() for blacklist_word in self.global_blacklist_words):
            return
        payload = MESG_regular.format(channel_url=channel_url, text=text, req_id=self.req_id)
        self.ws.send(payload)
        self.req_id += 1

    def ws_send_snoomoji(self, snoomoji, channel_url):
        if self.RateLimiter.is_enabled and self.RateLimiter.check():
            return
        payload = MESG_snoo.format(channel_url=channel_url, snoomoji=snoomoji, req_id=self.req_id)
        self.ws.send(payload)
        self.req_id += 1

    def ws_send_typing_indicator(self, channel_url):
        payload = TPST.format(channel_url=channel_url, time=int(time.time() * 1000))
        self.ws.send(payload)

    def ws_stop_typing_indicator(self, channel_url):
        payload = TPEN.format(channel_url=channel_url, time=int(time.time() * 1000))
        self.ws.send(payload)

    def on_error(self, ws, error):
        self.logger.error(error)
        self.last_err = error

    def on_close(self, ws):
        self.logger.warning("### websocket closed ###")



from .ws_client import WebSockClient
import pickle
from .reddit_auth import _RedditAuthBase, TokenAuth, PasswordAuth
from websocket import WebSocketConnectionClosedException
from ._api.tools import Tools
from ._api.models import Channel, Message
from typing import Dict, List, Callable, Optional
from ._utils.frame_model import FrameType, FrameModel

_hook = Callable[[FrameModel], Optional[bool]]


def _get_locals_without_self(locals_):
    del locals_['self']
    return locals_


class ChatBot:
    def __init__(self, authentication: _RedditAuthBase, store_session: bool = True, **kwargs):
        self._r_authentication = authentication
        self._store_session = store_session
        if store_session:
            if isinstance(authentication, TokenAuth):
                pkl_n = authentication.api_token
            elif isinstance(authentication, PasswordAuth):
                pkl_n = authentication.reddit_username
            else:
                pkl_n = None
            sb_access_token, user_id = self._load_session(pkl_n)
        else:
            reddit_authentication = self._r_authentication.authenticate()
            sb_access_token, user_id = reddit_authentication['sb_access_token'], reddit_authentication['user_id']

        self.WebSocketClient = WebSockClient(access_token=sb_access_token, user_id=user_id, **kwargs)
        self._tools = Tools(self._r_authentication)
        self.WebSocketClient._get_current_channels = self._tools.get_channels

    def get_chatroom_name_id_pairs(self) -> Dict[str, str]:
        return self.WebSocketClient.channelid_sub_pairs

    def on_message_hook(self, func: _hook) -> None:
        self.on_frame_hook(frame_type=FrameType.MESG)(func)

    def on_ready_hook(self, func: _hook) -> None:
        self.on_frame_hook(frame_type=FrameType.LOGI)(func)

    def on_frame_hook(self, frame_type: FrameType = FrameType.MESG) -> Callable[[_hook], None]:
        def on_frame_hook_append(func: _hook):
            def hook(resp: FrameModel):
                if resp.type_f == frame_type:
                    func(resp)
            self.WebSocketClient.after_message_hooks.append(hook)
        return on_frame_hook_append

    def set_respond_hook(self, input_: str,
                         response: str,
                         limited_to_users: List[str] = None,
                         lower_the_input: bool = False,
                         exclude_itself: bool = True,
                         must_be_equal: bool = True,
                         limited_to_channels: List[str] = None
                         ) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        if limited_to_users is None:
            limited_to_users = []
        try:
            response.format(nickname="")
        except KeyError as e:
            raise Exception("You need to set a {nickname} key in welcome message!") from e

        def hook(resp: FrameModel) -> Optional[bool]:
            sent_message = resp.message.lower() if lower_the_input else resp.message
            if (resp.user.name in limited_to_users or not bool(limited_to_users)) \
                    and (exclude_itself and resp.user.name != self.WebSocketClient.own_name) \
                    and ((must_be_equal and sent_message == input_) or (not must_be_equal and input_ in sent_message)) \
                    and (self.WebSocketClient.channelid_sub_pairs.get(
                resp.channel_url) in limited_to_channels or not bool(limited_to_channels)):
                response_prepped = response.format(nickname=resp.user.name)
                self.WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.on_frame_hook(frame_type=FrameType.MESG)(hook)

    def on_invitation_hook(self, func: _hook) -> None:
        def hook(resp: FrameModel) -> Optional[bool]:
            try:
                _ = resp.data.inviter
                invte = [invitee.nickname for invitee in resp.data.invitees]
            except AttributeError:
                return
            if not (len(invte) == 1 and invte[0] == self.WebSocketClient.own_name):
                return
            func(resp)

        self.on_frame_hook(frame_type=FrameType.SYEV)(hook)

    def on_message_deleted_hook(self, func: _hook) -> None:
        self.on_frame_hook(frame_type=FrameType.DELM)(func)

    def on_user_joined_hook(self, func: _hook):
        def hook(resp: FrameModel) -> Optional[bool]:
            try:
                _ = resp.data.users[0].nickname
                _ = resp.data.users[0].inviter.nickname
            except (AttributeError, IndexError):
                return
            func(resp)
        self.on_frame_hook(FrameType.SYEV)(hook)

    def set_welcome_message(self, message: str, limited_to_channels: List[str] = None) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        try:
            message.format(nickname="", inviter="")
        except KeyError as e:
            raise Exception("Keys should be {nickname} and {inviter}") from e

        def hook(resp: FrameModel) -> Optional[bool]:
            if self.WebSocketClient.channelid_sub_pairs.get(resp.channel_url) in limited_to_channels or \
                    not bool(limited_to_channels):
                response_prepped = message.format(nickname=resp.data.users[0].nickname,
                                                  inviter=resp.data.users[0].inviter.nickname)
                self.WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.on_user_joined_hook(hook)

    def on_user_left_hook(self, func: _hook) -> None:
        def hook(resp: FrameModel) -> Optional[bool]:
            try:
                _ = resp.channel.disappearing_message
                _ = resp.data.nickname
            except AttributeError:
                return
            func(resp)
        self.on_frame_hook(FrameType.SYEV)(hook)

    def set_farewell_message(self, message: str, limited_to_channels: List[str] = None) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        try:
            message.format(nickname="")
        except KeyError as e:
            raise Exception("Key should be {nickname}") from e

        def hook(resp: FrameModel) -> Optional[bool]:
            if self.WebSocketClient.channelid_sub_pairs.get(resp.channel_url) in limited_to_channels or \
                    not bool(limited_to_channels):
                response_prepped = message.format(nickname=resp.data.nickname)
                self.WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.on_user_left_hook(hook)

    def send_message(self, text: str, channel_url: str) -> None:
        self.WebSocketClient.ws_send_message(text, channel_url)

    def send_snoomoji(self, snoomoji: str, channel_url: str) -> None:
        self.WebSocketClient.ws_send_snoomoji(snoomoji, channel_url)

    def send_typing_indicator(self, channel_url: str) -> None:
        self.WebSocketClient.ws_send_typing_indicator(channel_url)

    def stop_typing_indicator(self, channel_url: str) -> None:
        self.WebSocketClient.ws_stop_typing_indicator(channel_url)

    def run_4ever(self, auto_reconnect: bool = True, max_retries: int = 500, skip_utf8_validation=True) -> None:
        for _ in range(max_retries):
            self.WebSocketClient.ws_run_forever(skip_utf8_validation=skip_utf8_validation)
            if self.WebSocketClient.is_logi_err and isinstance(self._r_authentication, PasswordAuth):
                self.WebSocketClient.logger.info('Re-Authenticating...')
                if self._store_session:
                    sb_access_token, _ = self._load_session(self._r_authentication.reddit_username, force_reauth=True)
                else:
                    sb_access_token = self._r_authentication.authenticate()['sb_access_token']
                self.WebSocketClient.update_ws_app_urls_access_token(sb_access_token)
            elif not (auto_reconnect and isinstance(self.WebSocketClient.last_err, WebSocketConnectionClosedException)):
                break
            self.WebSocketClient.logger.info('Auto Reconnecting...')

    def kick_user(self, channel_url: str, user_id: str, duration: int) -> None:
        self._tools.kick_user(**_get_locals_without_self(locals()))

    def delete_mesg(self, channel_url: str, msg_id: str) -> None:
        self._tools.delete_message(**_get_locals_without_self(locals()), session_key=self.WebSocketClient.session_key)

    def invite_user_to_channel(self, channel_url: str, nicknames: List[str]) -> None:
        self._tools.invite_user(**_get_locals_without_self(locals()))

    def get_chat_invites(self) -> List[Channel]:
        return self.get_channels(member_state_filter="invited_only")

    def get_channels(self, limit: int = 100, order: str = 'latest_last_message', show_member: bool = True,
                     show_read_receipt: bool = True, show_empty: bool = True, member_state_filter: str = 'joined_only',
                     super_mode: str = 'all', public_mode: str = 'all', unread_filter: str = 'all',
                     hidden_mode: str = 'unhidden_only', show_frozen: bool = True) -> List[Channel]:
        return self._tools.get_channels(**_get_locals_without_self(locals()),
                                        session_key=self.WebSocketClient.session_key)

    def get_older_messages(self, channel_url: str, prev_limit: int = 40, next_limit: int = 0, reverse: bool = True) -> List[Message]:
        return self._tools.get_older_messages(**_get_locals_without_self(locals()),
                                              session_key=self.WebSocketClient.session_key)

    def create_channel(self, nicknames: List[str], group_name: str) -> Channel:
        channel = self._tools.create_channel(**_get_locals_without_self(locals()),
                                             own_name=self.WebSocketClient.own_name)
        self.WebSocketClient.update_channelid_sub_pair()
        return channel

    def accept_chat_invite(self, channel_url: str) -> None:
        self._tools.accept_chat_invite(channel_url, session_key=self.WebSocketClient.session_key)
        self.WebSocketClient.update_channelid_sub_pair()

    def hide_chat(self, user_id: str, channel_url: str, hide_previous_messages: bool = False,
                  allow_auto_unhide: bool = True) -> None:
        self._tools.hide_chat(**_get_locals_without_self(locals()), session_key=self.WebSocketClient.session_key)

    def enable_rate_limiter(self, max_calls: float, period: float) -> None:
        self.WebSocketClient.RateLimiter.is_enabled = True
        self.WebSocketClient.RateLimiter.max_calls = max_calls
        self.WebSocketClient.RateLimiter.period = period

    def disable_rate_limiter(self) -> None:
        self.WebSocketClient.RateLimiter.is_enabled = False

    def _load_session(self, pkl_name, force_reauth=False):
        def get_store_file_handle(pkl_name_, mode_):
            try:
                return open(f'{pkl_name_}-stored.pkl', mode_)
            except FileNotFoundError:
                return None

        session_store_f = None if force_reauth else get_store_file_handle(pkl_name, 'rb')

        if session_store_f is None or force_reauth:
            session_store_f = get_store_file_handle(pkl_name, 'wb+')
            self._r_authentication.authenticate()
            pickle.dump(self._r_authentication, session_store_f)
        else:
            self._r_authentication = pickle.load(session_store_f)
        session_store_f.close()

        return self._r_authentication.sb_access_token, self._r_authentication.user_id



import time


class RateLimiter:
    is_enabled = False
    max_calls = 0
    period = 0
    _msg_counter = 0
    _period_end_ts = 0

    @staticmethod
    def check():
        current_ts = RateLimiter._get_current_ts()
        if RateLimiter._period_end_ts < current_ts:
            RateLimiter._msg_counter = 0
            RateLimiter._create_new_period(current_ts)
        if RateLimiter._msg_counter > RateLimiter.max_calls:
            return True
        else:
            RateLimiter._msg_counter += 1
            return False

    @staticmethod
    def _create_new_period(current_ts):
        RateLimiter._period_end_ts = current_ts + RateLimiter.period * 60

    @staticmethod
    def _get_current_ts():
        return int(time.time())



from setuptools import setup

setup(
    name='Reddit_ChatBot_Python',
    version='1.2.6',
    project_urls={"Source": "https://github.com/scrubjay55/Reddit_ChatBot_Python.git"},
    url="https://github.com/scrubjay55/Reddit_ChatBot_Python.git",
    author='scrubjay55',
    license="APLv2",
    packages=['Reddit_ChatBot_Python', 'Reddit_ChatBot_Python._utils', 'Reddit_ChatBot_Python._api'],
    description='an event-driven chatbot library for reddit chatrooms',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "requests",
        "websocket_client==0.58.0",
        "dataclasses-json"
    ],
    extras_require={"extra": ["numpy", "wsaccel"]}
)



import requests
import uuid
from ._utils.consts import MOBILE_USERAGENT, OAUTH_REDDIT, S_REDDIT, WEB_USERAGENT, OAUTH_CLIENT_ID_B64, WWW_REDDIT, ACCOUNTS_REDDIT


class _RedditAuthBase:
    def __init__(self, api_token=None):
        self.api_token = api_token
        self.sb_access_token = None
        self.user_id = None

    def authenticate(self):
        self._get_userid_sb_token()
        return {'sb_access_token': self.sb_access_token, 'user_id': self.user_id}

    def _get_userid_sb_token(self):
        headers = {
            'User-Agent': MOBILE_USERAGENT,
            'Authorization': f'Bearer {self.api_token}'
        }
        sb_token_j = requests.get(f'{S_REDDIT}/api/v1/sendbird/me', headers=headers).json()
        self.sb_access_token = sb_token_j['sb_access_token']
        user_id_j = requests.get(f'{OAUTH_REDDIT}/api/v1/me.json', headers=headers).json()
        self.user_id = 't2_' + user_id_j['id']


class PasswordAuth(_RedditAuthBase):
    def __init__(self, reddit_username: str, reddit_password: str, twofa: str = None):
        super().__init__()
        self.reddit_username = reddit_username
        self.reddit_password = reddit_password
        self.twofa = twofa
        self._client_vendor_uuid = str(uuid.uuid4())
        self._reddit_session = None

    def authenticate(self):
        if not (self._reddit_session is None and self.api_token is None):
            self.refresh_api_token()
        else:
            self._reddit_session = self._do_login()
            if self._reddit_session is None:
                raise Exception("Wrong username or password")
            self.api_token = self._get_api_token(self._reddit_session)

        return super(PasswordAuth, self).authenticate()

    def _get_api_token(self, reddit_session):
        cookies = {'reddit_session': reddit_session}
        headers = {
            'Authorization': f'Basic {OAUTH_CLIENT_ID_B64}',
            'User-Agent': MOBILE_USERAGENT,
            'client-vendor-id': self._client_vendor_uuid,
        }
        data = '{"scopes":["*"]}'
        response = requests.post(f'{ACCOUNTS_REDDIT}/api/access_token', headers=headers, cookies=cookies, data=data).json()
        api_token = response['access_token']
        return api_token

    def refresh_api_token(self):
        assert self._reddit_session is not None
        cookies = {
            'reddit_session': self._reddit_session,
        }
        headers = {
            'User-Agent': WEB_USERAGENT,
            'Authorization': f'Bearer {self.api_token}',
        }
        data = {
            'accessToken': self.api_token,
            'unsafeLoggedOut': 'false',
            'safe': 'true'
        }
        response = requests.post(f'{WWW_REDDIT}/refreshproxy', headers=headers, cookies=cookies, data=data).json()
        self.api_token = response['accessToken']
        return self.api_token

    def _do_login(self):
        headers = {
            'User-Agent': WEB_USERAGENT,
        }
        data = {
            'op': 'login',
            'user': self.reddit_username,
            'passwd': f'{self.reddit_password}:{self.twofa}' if bool(self.twofa) else self.reddit_password,
            'api_type': 'json'
        }
        response = requests.post(f'{WWW_REDDIT}/api/login/{self.reddit_username}', headers=headers, data=data)
        reddit_session = response.cookies.get("reddit_session")
        return reddit_session


class TokenAuth(_RedditAuthBase):
    def __init__(self, token):
        super().__init__(token)



from urllib.parse import urlencode
from .consts import SB_User_Agent, SB_ai


def get_ws_url(user_id, access_token):
    socket_base = "wss://sendbirdproxyk8s.chat.redditmedia.com"
    ws_params = {
        "user_id": user_id,
        "access_token": access_token,
        "p": "Android",
        "pv": 30,
        "sv": "3.0.144",
        "ai": SB_ai,
        "SB-User-Agent": SB_User_Agent,
        "active": "1"
    }
    return f"{socket_base}/?{urlencode(ws_params)}"


def pair_channel_and_names(channels: list, user_id: str):
    channelid_sub_pairs = {}
    for channel in channels:
        room_name = None
        if channel.custom_type == "direct":
            for member in channel.members:
                if member.user_id != user_id:
                    room_name = member.nickname
                    break
        else:
            room_name = channel.channel.name
        channelid_sub_pairs.update({channel.channel.channel_url: room_name})
    return channelid_sub_pairs
